from contextvars import ContextVar
import datetime
import inspect
import logging
import multiprocessing
import threading
import time
import traceback
import uuid
from contextlib import AbstractContextManager
from typing import Any, Dict, Iterable, Optional, Type, List, Callable

from oplog.exceptions import (
    GlobalOperationPropertyAlreadyExistsException,
    OperationPropertyAlreadyExistsException
)
from oplog.operation_step import OperationStep

active_operation_stack: ContextVar[List['Operation']] = (
    ContextVar("active_operation_stack", default=[])
)


class Operation(AbstractContextManager):
    global_props: Dict[str, Any] = {}
    _serializer: Optional[Callable[['Operation'], str]] = None

    @classmethod
    def config(cls, serializer: Callable[['Operation'], str]) -> None:
        """
        Configure global behavior of all operations.

        :param serializer: A function that receives an operation
        and returns a string, to be formatted where relevant.
        :return:
        """
        # Any attributes that are set here should be cleaned in `factory_reset`
        cls._serializer = serializer

    def __init__(self,
                 name: str,
                 suppress: bool = False,
                 on_start: bool = False) -> None:
        """
        Initialize a new operation, to be used as a context manager,
        and create a context-rich operation log, to be handled by
         `oplog.handlers.OperationLogHandler`.
        :param name: The name of the operation.
        :param suppress: if True, the context manager will suppress
        any exception thrown in the context. Otherwise, the exception
        will be raised. Default is False. Operation metadata will be set properly
        regardless of this flag (e.g., failed operation).
        :param on_start: if True, the operation will be logged on start
        (as well as on exit).
        """
        # Check if there's an active operation and assign parent-child relationship
        self.parent_op: Optional[Operation] = None
        self.child_ops: List[Operation] = []
        if active_operation_stack.get() and active_operation_stack.get()[-1]:
            self.parent_op = active_operation_stack.get()[-1]
            self.parent_op.child_ops.append(self)

        self.name = name
        self.suppress = suppress
        self._on_start = on_start
        self.custom_props: Dict[str, Any] = dict()

        self.start_time_utc: Optional[datetime.datetime] = None
        self.end_time_utc: Optional[datetime.datetime] = None
        self.start_time_utc_str: Optional[str] = None
        self.end_time_utc_str: Optional[str] = None
        self.duration_ms: Optional[int] = None
        self.duration_ns: Optional[int] = None
        self.id = str(uuid.uuid4())
        self.step: Optional[OperationStep] = None
        self.is_successful: Optional[bool] = None
        self.result: Optional[str] = None
        self.exception_type: Optional[str] = None
        self.exception_msg: Optional[str] = None

        self._logger = self._get_caller_logger()
        self.logger_name = self._logger.name
        self.log_level: Optional[str] = None

        current_proc = multiprocessing.current_process()
        self.process_name = current_proc.name
        self.process_id = current_proc.pid
        self.thread_name = threading.current_thread().name
        self.thread_id = threading.get_ident()

        # inheritable props
        self.correlation_id: Optional[str] = None

        self._perf_start: Optional[float] = None


    @classmethod
    def factory_reset(cls) -> None:
        cls.global_props = {}
        cls._serializer = None

    @classmethod
    def _get_caller_logger(cls) -> logging.Logger:
        logger = None
        stack = inspect.stack()
        if len(stack) >= 3:
            # the caller is the 3rd frame in the stack
            caller_frame = stack[2]
            if len(caller_frame) >= 1:
                # the caller module is the 1st element in the caller frame
                caller_module = inspect.getmodule(caller_frame[0])
                if caller_module is not None:
                    logger = logging.getLogger(caller_module.__name__)

        if logger is None:
            logger = logging.getLogger()
        return logger

    def __str__(self):
        if self._serializer:
            return self.__class__._serializer(self)

        msg = (f"{self.start_time_utc_str} ({self.duration_ms}ms): "
               f"[{self.name} / {self.result}]")
        return msg

    def __enter__(self) -> "Operation":
        self.start_time_utc = datetime.datetime.utcnow()
        # time format example: 2023-06-22 06:27:53.922633
        self.start_time_utc_str = self.start_time_utc.strftime("%Y-%m-%d %H:%M:%S.%f")
        self._perf_start = time.perf_counter_ns()

        self.step = OperationStep.START

        self.set_inheritable_props()

        # Push the current operation onto the stack
        active_operation_stack.set(active_operation_stack.get([]) + [self])

        if self._on_start:
            self._logger.log(
                level=logging.INFO,
                msg=str(self),
                extra={"oplog": self}
            )

        return self

    def set_inheritable_props(self) -> None:
        if self.parent_op is not None:
            self.correlation_id = self.parent_op.correlation_id
        else:
            self.correlation_id = str(uuid.uuid4())

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.end_time_utc = datetime.datetime.utcnow()
        self.end_time_utc_str = self.end_time_utc.strftime("%Y-%m-%d %H:%M:%S.%f")

        perf_end = time.perf_counter_ns()
        self.duration_ms = round((perf_end - self._perf_start) / 1_000_000)
        self.duration_ns = round(perf_end - self._perf_start)

        self.step = OperationStep.END

        # Pop the current operation off the stack
        current_stack = active_operation_stack.get([])
        if current_stack:
            current_stack.pop()
            active_operation_stack.set(current_stack)

        is_success = exc_type is None
        if is_success:
            result = "Success"
            tb = ""
        else:
            result = "Failure"
            self.exception_type = exc_type.__name__
            self.exception_msg = str(exc_value)
            tb = traceback.extract_tb(exc_tb, limit=10).format()

        self.is_successful = is_success
        self.result = result
        self.traceback = tb
        level = logging.INFO if is_success else logging.ERROR
        self.log_level = logging.getLevelName(level)

        self._logger.log(
            level=level,
            msg=str(self),
            extra={"oplog": self}
        )

        # this will either suppress (if configured) or no,
        # in case an error was thrown in context
        return self.suppress

    def __hash__(self):  # pragma: no cover
        return hash(self.id)

    def __eq__(self, other):  # pragma: no cover
        if isinstance(other, Operation):
            return other.id == self.id
        return False

    def add(self, prop_name: str, value: Any) -> None:
        self._add_custom_prop(property_name=prop_name, value=value)

    @classmethod
    def add_global(cls, prop_name: str, value: Any) -> None:
        cls._validate_type(value=value, expected_types=(str, int, float, bool))
        cls._add_global_prop(property_name=prop_name, value=value)

    @staticmethod
    def _validate_type(value: Any, expected_types: Iterable[Type]):
        if not any(isinstance(value, t) for t in expected_types):
            expected_types_str = " or ".join(str(t) for t in expected_types)
            raise TypeError(f"Expected {expected_types_str}, but got {type(value)}")

    def add_multi(self, bag: Dict[str, Any]) -> None:
        for prop_name in bag:
            self.add(prop_name=prop_name, value=bag[prop_name])

    def _add_custom_prop(self, property_name: str, value: Any) -> None:
        if property_name in self.custom_props:
            raise OperationPropertyAlreadyExistsException(
                op_name=self.name, prop_name=property_name
            )
        self.custom_props[property_name] = value

    @classmethod
    def _add_global_prop(cls, property_name: str, value: Any) -> None:
        if property_name in cls.global_props:
            raise GlobalOperationPropertyAlreadyExistsException(prop_name=property_name)
        cls.global_props[property_name] = value

    def __repr__(self):  # pragma: no cover
        return f"<Operation name={self.name}>"




from oplog import Operation
from oplog.formatters import VerboseOperationFormatter
from oplog.tests.logged_test_case import OpLogTestCase


class VerboseOplogLineFormatterTestException(Exception):
    pass


class TestVerboseOperationFormatter(OpLogTestCase):
    def test_formatOp_operationWithCustomProperties(self):
        # arrange
        op_name = "test_operation"
        custom_prop_name = "some_custom_prop"
        custom_prop_value = "some_custom_value"
        with Operation(name=op_name) as op:
            op.add(custom_prop_name, custom_prop_value)
        formatter = VerboseOperationFormatter()

        # act
        log_line = formatter.format_op(op=op)

        # assert
        self.assertIn(custom_prop_name, log_line)
        self.assertIn(custom_prop_value, log_line)

    def test_formatOp_operationWithGlobalProperties(self):
        # arrange
        op_name = "test_operation"
        global_prop_name = "some_global_prop"
        global_prop_value = "some_global_value"
        with Operation(name=op_name) as op:
            op.add_global(global_prop_name, global_prop_value)
        formatter = VerboseOperationFormatter()

        # act
        log_line = formatter.format_op(op=op)

        # assert
        self.assertIn(global_prop_name, log_line)
        self.assertIn(global_prop_value, log_line)

    def test_formatOp_operationFailure(self):
        # arrange
        op_name = "test_operation"
        error_msg = "some error message"
        with Operation(name=op_name, suppress=True) as op:
            raise VerboseOplogLineFormatterTestException(error_msg)
        formatter = VerboseOperationFormatter()

        # act
        log_line = formatter.format_op(op=op)

        # assert
        self.assertIn(VerboseOplogLineFormatterTestException.__name__, log_line)
        self.assertIn(error_msg, log_line)



from oplog.formatters.base_operation_formatter import BaseOperationFormatter
from oplog.operation import Operation


class CsvOperationFormatter(BaseOperationFormatter):
    def format_op(self, op: Operation) -> str:
        csv_row = [
            f'"{op.start_time_utc_str}"',
            f'"{str(op.duration_ms)}"',
            f'"{op.name}"',
            f'"{op.correlation_id}"',
            f'"{op.result}"',
            f'"{str(op.exception_type)}"',
        ]
        return ','.join(csv_row)



import pytest
import os
import sys

current_dir = os.path.dirname(os.path.abspath(__file__))
# Get the path to the repository root directory (one level up)
repository_root = os.path.abspath(os.path.join(current_dir, "..", ".."))

# Add the repository root directory to the Python path
sys.path.append(repository_root)


from examples.fluent_calculator_demo.main import FluentCalculator  # noqa: E402


@pytest.fixture
def calculator():
    return FluentCalculator()


def test_add(calculator):
    # act
    result = calculator.add(5).add(3).add(2).calc()
    # assert
    assert result == 10



import csv

from oplog import Operation
from oplog.formatters import CsvOperationFormatter
from oplog.tests.logged_test_case import OpLogTestCase


class TestCsvOperationFormatter(OpLogTestCase):
    def test_formatOp_operationNameContainsSeparator_internalSeparatorIsEscaped(self):
        # arrange
        separator = ","
        expected_num_columns = (
            self._get_num_columns_for_csv_formatted_operation("operation_name"))

        # act
        actual_num_columns = (
            self._get_num_columns_for_csv_formatted_operation(f"operation{separator}name"))

        # assert
        self.assertEqual(expected_num_columns, actual_num_columns)

    @staticmethod
    def _get_num_columns_for_csv_formatted_operation(operation_name: str) -> int:
        op = Operation(name=operation_name)
        formatter = CsvOperationFormatter()
        csv_str = formatter.format_op(op=op)
        csv_reader = csv.reader(csv_str.splitlines(), delimiter=',')
        first_row = next(csv_reader)
        num_columns = len(first_row)
        return num_columns



from .csv_operation_formatter import CsvOperationFormatter  # noqa: F401
from .verbose_operation_formatter import VerboseOperationFormatter  # noqa: F401
from .base_operation_formatter import BaseOperationFormatter  # noqa: F401



import logging
from unittest.mock import Mock, patch

from oplog import Operation
from oplog.exceptions import LogRecordMissingOperationException
from oplog.formatters import BaseOperationFormatter
from oplog.tests.logged_test_case import OpLogTestCase


class MyTestOperationFormatter(BaseOperationFormatter):
    def format_op(self, op: Operation) -> str:
        return op.name


class TestBaseOperationFormatter(OpLogTestCase):
    def test_format_logRecordIsNotOplog_errorRaised(self):
        # arrange
        mock_record = Mock(spec=logging.LogRecord)
        formatter = MyTestOperationFormatter()

        # act
        with self.assertRaises(LogRecordMissingOperationException):
            formatter.format(record=mock_record)

    def test_format_callsFormatOp(self):
        # arrange
        formatter = MyTestOperationFormatter()
        op_name = "test_operation"
        mock_record = Mock(spec=logging.LogRecord)
        with patch.object(
                MyTestOperationFormatter,
                MyTestOperationFormatter.format_op.__name__,
                return_value=None) as format_op_mock:
            with Operation(name=op_name) as op:
                mock_record.oplog = op

            # act
            formatter.format(record=mock_record)

            # assert
            format_op_mock.assert_called_once_with(op=mock_record.oplog)




from oplog.operation_step import OperationStep

from oplog.formatters.base_operation_formatter import BaseOperationFormatter
from oplog.operation import Operation


class VerboseOperationFormatter(BaseOperationFormatter):
    def format_op(self, op: Operation) -> str:
        duration = f" ({op.duration_ms}ms)" if op.step == OperationStep.END else ""
        result = op.result if op.result else "started"
        msg = (f"{op.start_time_utc_str}{duration}: "
               f"[{op.name} / {result}]")
        if op.exception_type:
            msg += f" {op.exception_type}: {op.exception_msg}"

        if len(op.custom_props) > 0:
            msg += f" {op.custom_props}"

        if len(op.global_props) > 0:
            msg += f" {op.global_props}"

        return msg



import logging
from abc import abstractmethod, ABC

from oplog.exceptions import LogRecordMissingOperationException
from oplog.operation import Operation


class BaseOperationFormatter(logging.Formatter, ABC):
    def format(self, record) -> str:
        if hasattr(record, "oplog"):
            op: Operation = record.oplog
            return self.format_op(op=op)
        else:
            raise LogRecordMissingOperationException(
                "record does not have an oplog property"
            )

    @abstractmethod
    def format_op(self, op: Operation) -> str:  # pragma: no cover
        raise NotImplementedError("format_operation is required")
